--------------------------------------------------------------------------
-- AISYSTEM
--
-- Copyright (C) 2007-Present Inventikon Corporation. all rights reserved.
--
-- See LICENSE.TXT document for licensing information.
--------------------------------------------------------------------------
library ieee;
use ieee.std_logic_1164.all;
use ieee.numeric_std.all;

library ieee_proposed;
use ieee_proposed.fixed_float_types.all;
use ieee_proposed.fixed_pkg.all;
use ieee_proposed.float_pkg.all;

library fplib;
use fplib.pkg_fplib_std.all;

library work;
use work.common.all;

library work;
use work.definition.all;

-- in order to integrate a first order differential equation using the function(s) below 
-- we need to transform it into:
-- 		y' = a - b * y

entity integration is
    port
    (
		reset           : in std_logic;
		clock           : in std_logic;
		-- inputs
		request         : in std_logic;
		dt              : in float(WE downto -WF);
		y               : in float(WE downto -WF);
		a               : in float(WE downto -WF);
		b               : in float(WE downto -WF);
		-- outputs
		result          : out float(WE downto -WF);
		acknowledge		: out std_logic
    );
end entity;


architecture integration_runge_kutta4 of integration is
	type state_type is (INITIALIZATION, CALCULATION);

	signal state            : state_type;
	signal dt_f				: float(WE downto -WF);
	signal minus_b     		: float(WE downto -WF);
	signal temp1			: float(WE downto -WF);
	signal temp2			: float(WE downto -WF);
	signal temp3			: float(WE downto -WF);
	signal temp4			: float(WE downto -WF);
	signal k1				: float(WE downto -WF);
	signal k2				: float(WE downto -WF);
	signal k3				: float(WE downto -WF);
	signal k4				: float(WE downto -WF);
	signal step             : integer range 0 to 31 := 0;
	signal ack_int          : std_logic := '0';

	-- synthesis translate_off
	-- signals used for debug purposes only (i.e. not included in synthesis)
	signal result1_real, result2_real, result3_real, result4_real : real;
	signal k1_real, k2_real, k3_real, k4_real : real;
	-- synthesis translate_on
begin
    
    dt_f <= dt;
    acknowledge <= ack_int;
    
    -- synthesis translate_off
    result1_real <= to_real(temp1);
    result2_real <= to_real(temp2);
    result3_real <= to_real(temp3);
    result4_real <= to_real(temp4);
    k1_real <= to_real(k1);
    k2_real <= to_real(k2);
    k3_real <= to_real(k3);
    k4_real <= to_real(k4);
    -- synthesis translate_on
    
	sm_integration : process (reset, clock) is
    begin
        if reset = '1' then
            state   <= INITIALIZATION;
            result  <= (others=>'0');
            step    <= 0;
        elsif rising_edge(clock) then
            ack_int <= '0';
			
            case state is
                when INITIALIZATION =>
					if request = '1' and ack_int = '0' then
                        state <= CALCULATION;
                    else
                        state <= INITIALIZATION;
                    end if;
                when CALCULATION =>    
                    step <= step + 1;
                    state <= CALCULATION;
                    case step is
                        when 0 =>
                            temp1 <= b * y;           				-- b * yn
                        when 1 =>
                            temp2 <= a - temp1;       				-- (a - b * yn)
                        when 2 => 
                            k1 <= temp2 * dt_f;  					-- k1 = (a - b * yn) * dt
                        when 3 =>
                            temp1 <= k1 / to_float(2, temp1); 		-- k1 / 2
                        when 4 =>     
                            temp2 <= y + temp1;     				-- (yn + k1 / 2)
                        when 5 =>     
                            temp3 <= b * temp2;     				-- b*(yn + k1 / 2)
                        when 6 =>
                            temp1 <= a - temp3;     				-- (a - b * (yn + k1 / 2))
                        when 7 =>
                            k2 <= temp1 * dt_f;  					-- k2 = (a - b * (yn + k1 / 2)) * dt
                        when 8 =>
                            temp3 <= k2 / to_float(2, temp1); 		-- k2 / 2
                        when 9 =>     
                            temp1 <= y + temp3;     				-- (yn + k2 / 2)
                        when 10 =>     
                            temp2 <= b * temp1;     				-- b*(yn + k2 / 2)
                        when 11 =>
                            temp3 <= a - temp2;     				-- (a - b * (yn + k2 / 2))
                        when 12 =>
                            k3 <= temp3 * dt_f;  					-- k3 = (a - b * (yn + k2 / 2)) * dt  
                        when 13 => 
                            temp2 <= y + k3;     					-- (yn + k3)
                        when 14 =>     
                            temp3 <= b * temp2;     				-- b * (yn + k3)
                        when 15 => 
                            temp1 <= a - temp3;     				-- (a - b * (yn + k3))
                        when 16 =>    
                            k4 <= temp1 * dt_f;     				-- k4 = (a - b * (yn + k3)) * dt
                        when 17 =>
                            temp2 <= k2 + k3;                   	-- k2 + k3
                            temp3 <= k1 + k4;                   	-- k1 + k4
                        when 18 => 
                            temp1 <= to_float(2, temp1) * temp2; 	-- 2 * k2 + 2 * k3
                        when 19 =>  
                            temp2 <= temp1 + temp3;             	-- (k1 + 2 * k2 + 2 * k3 + k4)
                        when 20 =>
                            temp1 <= temp2 / to_float(6, temp1);    -- (k1 + 2 * k2 + 2 * k3 + k4) / 6
                        when 21 => 
                            result <= y + temp1;  					-- yn1 = yn + (k1 + 2 * k2 + 2 * k3 + k4) / 6
                            state <= INITIALIZATION;
                            step <= 0;
                            ack_int <= '1';
                        when others => 
                            state <= INITIALIZATION;
                            step <= 0;
                    end case;        
            end case;        
        end if;
    end process;
end architecture;


architecture integration_forward_euler of integration is
	type state_type is (INITIALIZATION, CALCULATION);

	signal state            : state_type;
	signal dt_f             : float(WE downto -WF);
	signal temp1			: float(WE downto -WF);
	signal temp2 			: float(WE downto -WF);
	signal temp3		 	: float(WE downto -WF);
	signal temp4		 	: float(WE downto -WF);
	signal step             : integer range 0 to 31 := 0;
	signal ack_int          : std_logic := '0';

	-- synthesis translate_off
	-- signals used for debug purposes only (i.e. not included in synthesis)
	signal result1_real, result2_real, result3_real, result4_real : real;
	-- synthesis translate_on
begin  
    dt_f <= dt;
    acknowledge <= ack_int;
    
    -- synthesis translate_off
    result1_real <= to_real(temp1);
    result2_real <= to_real(temp2);
    result3_real <= to_real(temp3);
    result4_real <= to_real(temp4);
    -- synthesis translate_on
    
	sm_integration : process (reset, clock) is
    begin
        if reset = '1' then
            state <= INITIALIZATION;
            result <= (others=>'0');
            step <= 0;
        elsif rising_edge(clock) then
            ack_int <= '0';
            case state is
                when INITIALIZATION =>
                    if request = '1' and ack_int = '0' then
                        state <= CALCULATION;
                    else
                        state <= INITIALIZATION;
                    end if;
                when CALCULATION =>
                    step <= step + 1;
                    state <= CALCULATION;
                    case step is
                        when 0 =>
                            temp1 <= b * y;             -- b * yn         
                        when 1 =>
                            temp2 <= a - temp1;         -- (a - b * yn)
                        when 2 => 
                            temp3 <= temp2 * dt_f;		-- (a - b * yn) * dt
                        when 3 =>
                            result <= y + temp3;        -- yn1 = yn + (a - b * yn) * dt
                            state <= INITIALIZATION;
                            step <= 0;
                            ack_int <= '1';
                        when others => 
                            state <= INITIALIZATION;
                            step <= 0;
                    end case;        
            end case;        
        end if;
    end process;
end architecture;


architecture integration_exponential_euler of integration is
	type state_type is (INITIALIZATION, CALCULATION);
	
	signal state            : state_type;
	signal dt_f				: float(WE downto -WF);
	signal minus_b    		: float(WE downto -WF);
	signal temp1			: float(WE downto -WF);
	signal temp2 			: float(WE downto -WF);
	signal temp3		 	: float(WE downto -WF);
	signal temp4		 	: float(WE downto -WF);
	signal step             : integer range 0 to 31 := 0;
	signal ack_int          : std_logic := '0';
	-- fplib signals
	signal exp_in			: std_logic_vector(WEF + 2 downto 0);
	signal exp_result 		: std_logic_vector(WEF + 2 downto 0);

	-- synthesis translate_off
	-- signals used for debug purposes only (i.e. not included in synthesis)
	signal result1_real, result2_real, result3_real, result4_real : real;
	-- synthesis translate_on
begin
    unit_exp : entity fplib.exp_clk
    generic map (WE  => WE,
                 WF  => WF,
                 reg => false,
                 SzSlk => 0 )
    port map (na  => exp_in,
              nr  => exp_result,
              clk => clock);
    
    dt_f <= dt;
    acknowledge <= ack_int;
    
    -- synthesis translate_off
    result1_real <= to_real(temp1);
    result2_real <= to_real(temp2);
    result3_real <= to_real(temp3);
    result4_real <= to_real(temp4);
    -- synthesis translate_on
    
	sm_integration : process (reset, clock) is
    begin
        if reset = '1' then
            state   <= INITIALIZATION;
            result  <= (others=>'0');
            step    <= 0;
        elsif rising_edge(clock) then
            ack_int <= '0';
            case state is
                when INITIALIZATION =>
                    if request = '1' and ack_int = '0' then
                        state <= CALCULATION;
                    else
                        state <= INITIALIZATION;
                    end if;
                    minus_b <= b;
                    minus_b(WE) <= not b(WE); 							-- changing the sign of b
                when CALCULATION =>
                    step <= step+1;
                    state <= CALCULATION;
                    case step is
                        when 0 =>   
                            temp4 <= minus_b * dt_f;              		-- -b * dt                    
                        when 1 =>
                            exp_in(WEF + 2 downto WEF + 1) <= b"01";

                            exp_in(WEF downto 0) <= to_slv(temp4); 		-- D = exp(-b * dt)    
                        when 2 =>
                            temp1 <= a / b;                       		-- (a / b)
                        when 3 =>
                            temp4 <= to_float(exp_result(WEF downto 0), temp1);
                        when 4 => 
                            temp2 <= y * temp4;                 		-- yn * D
                            temp3 <= to_float(1, temp1) - temp4;   		-- (1 - D)
                        when 5 =>    
                            temp4 <= temp1 * temp3;             		-- (a / b) * (1 - D)
                        when 6 =>    
                            result <= temp2 + temp4;              		-- yn1 = yn * D + (a / b) * (1 - D)
                            state <= INITIALIZATION;
                            step <= 0;
                            ack_int <= '1';
                        when others =>
                            state <= INITIALIZATION;
                            step <= 0;
                    end case;        
            end case;        
        end if;
    end process;
end architecture;


architecture integration_backward_euler of integration is
	type state_type is (INITIALIZATION, CALCULATION);

	signal state            : state_type;
	signal dt_f             : float(WE downto -WF);
	signal temp1			: float(WE downto -WF);
	signal temp2 			: float(WE downto -WF);
	signal temp3		 	: float(WE downto -WF);
	signal temp4		 	: float(WE downto -WF);
	signal step             : integer range 0 to 31 := 0;
	signal ack_int          : std_logic := '0';

	-- synthesis translate_off
	-- signals used for debug purposes only (i.e. not included in synthesis)
	signal result1_real, result2_real, result3_real, result4_real : real;
	-- synthesis translate_on
begin
    dt_f <= dt;
    acknowledge <= ack_int;
    
    -- synthesis translate_off
    result1_real <= to_real(temp1);
    result2_real <= to_real(temp2);
    result3_real <= to_real(temp3);
    result4_real <= to_real(temp4);
    -- synthesis translate_on
    
	sm_integration : process (reset, clock) is
    begin
        if reset = '1' then
            state   <= INITIALIZATION;
            result  <= (others=>'0');
            step    <= 0;
        elsif rising_edge(clock) then
            ack_int <= '0';
            case state is
                when INITIALIZATION =>
                    if request = '1' and ack_int = '0' then
                        state <= CALCULATION;
                    else
                        state <= INITIALIZATION;
                    end if;
                when CALCULATION =>
                    step <= step + 1;
                    state <= CALCULATION;
                    case step is
                        when 0 =>
                            temp1 <= a * dt_f;                    	--  a * dt
                            temp2 <= b * dt_f;                    	--  b * dt                          
                        when 1 =>
                            temp1 <= y + temp1;                   	-- (yn + a * dt)
                            temp2 <= to_float(1, temp2) + temp2;  	-- (1 + b * dt)
                        when 2 => 
                            result <= temp1 / temp2;              	-- yn1 = (yn + a * dt) / (1 + b * dt)
                            state <= INITIALIZATION;
                            step <= 0;
                            ack_int <= '1';
                        when others => 
                            state <= INITIALIZATION;
                            step <= 0;
                    end case;
            end case;
        end if;
    end process;
end architecture;


architecture integration_trapezoid of integration is
	type state_type is (INITIALIZATION, CALCULATION);

	signal state            : state_type;
	signal dt_f             : float(WE downto -WF);
	signal temp1			: float(WE downto -WF);
	signal temp2 			: float(WE downto -WF);
	signal temp3		 	: float(WE downto -WF);
	signal temp4		 	: float(WE downto -WF);
	signal step             : integer range 0 to 31 := 0;
	signal ack_int          : std_logic := '0';

	-- synthesis translate_off
	-- signals used for debug purposes only (i.e. not included in synthesis)
	signal result1_real, result2_real, result3_real, result4_real : real;
	-- synthesis translate_on
begin
    dt_f <= dt;
    acknowledge <= ack_int;
    
    -- synthesis translate_off
    result1_real <= to_real(temp1);
    result2_real <= to_real(temp2);
    result3_real <= to_real(temp3);
    result4_real <= to_real(temp4);
    -- synthesis translate_on
    
	sm_integration : process (reset, clock) is
    begin
        if reset = '1' then
            state   <= INITIALIZATION;
            result  <= (others=>'0');
            step    <= 0;
        elsif rising_edge(clock) then
            ack_int <= '0';
            case state is
                when INITIALIZATION =>
                    if request = '1' and ack_int = '0' then
                        state <= CALCULATION;
                    else
                        state <= INITIALIZATION;
                    end if;
                when CALCULATION => 
                    step <= step + 1;
                    state <= CALCULATION;
                    case step is
                        when 0 =>
                            temp2 <= dt_f * b;                    -- dt * b
                        when 1 =>
                            temp3 <= temp2 / to_float(2, temp1);  -- dt * b / 2.0 
                        when 2 =>
                            temp1 <= to_float(1, temp1) - temp3;  -- (1 - dt * b / 2.0)
                            temp2 <= to_float(1, temp1) + temp3;  -- (1 + dt * b / 2.0)
                            temp3 <= dt_f * a;                    -- dt * a
                        when 3 =>
                            temp1 <= y * temp1;                   -- yn * (1 - dt * b / 2.0)
                        when 4 =>
                            temp4 <= temp1 + temp3;               -- (yn * (1 - dt * b / 2.0) + dt * a)
                        when 5 =>
                            result <= temp4 / temp2;              -- yn1 = (yn * (1 - dt * b / 2.0) + dt * a) / (1 + dt * b / 2.0)
                            state <= INITIALIZATION;
                            step <= 0;
                            ack_int <= '1';
                        when others => 
                            state <= INITIALIZATION;
                            step <= 0;
                    end case;   
            end case;        
        end if;
    end process;
end architecture;


architecture integration_crank_nicholson of integration is
	type state_type is (INITIALIZATION, CALCULATION);

	signal state            : state_type;
	signal dt_f             : float(WE downto -WF);
	signal temp1			: float(WE downto -WF);
	signal temp2 			: float(WE downto -WF);
	signal temp3		 	: float(WE downto -WF);
	signal temp4		 	: float(WE downto -WF);
	signal step             : integer range 0 to 31 := 0;
	signal ack_int          : std_logic := '0';

	-- synthesis translate_off
	-- signals used for debug purposes only (i.e. not included in synthesis)
	signal result1_real, result2_real, result3_real, result4_real : real;
	-- synthesis translate_on
begin
    
    dt_f <= dt;
    acknowledge <= ack_int;
    
    -- synthesis translate_off
    result1_real <= to_real(temp1);
    result2_real <= to_real(temp2);
    result3_real <= to_real(temp3);
    result4_real <= to_real(temp4);
    -- synthesis translate_on
    
	sm_integration : process (reset, clock) is
    begin
        if reset = '1' then
            state   <= INITIALIZATION;
            result  <= (others=>'0');
            step    <= 0;
        elsif rising_edge(clock) then
            ack_int <= '0';
            case state is
                when INITIALIZATION =>
                    if request = '1' and ack_int = '0' then
                        state <= CALCULATION;
                    else
                        state <= INITIALIZATION;
                    end if;
                when CALCULATION => 
                    step <= step + 1;
                    state <= CALCULATION;
                    case step is
                        when 0 =>
                            temp1 <= a * dt_f;                  	-- a * dt
                            temp2 <= b * dt_f;                  	-- b * dt
                        when 1 =>
                            temp3 <= to_float(1, temp1) + temp2;	-- (1 + b * dt)
                            temp4 <= y + temp1;                 	-- (yn + a * dt)
                        when 2 =>
                            temp1 <= temp4 / temp3;               	-- yhalf = (yn + a * dt) / (1 + b * dt)
                        when 3 =>
                            temp2 <= to_float(2, temp1) * temp1; 	-- 2 * yhalf
                        when 4 =>
                            result <= temp2 - y;                  	-- yn1 = 2 * yhalf - yn
                            state <= INITIALIZATION;
                            step <= 0;
                            ack_int <= '1';
                        when others => 
                            state <= INITIALIZATION;
                            step <= 0;
                    end case;   
                when others =>
                    null;
            end case;        
        end if;
    end process;
end architecture;

