------------------------------------------------------------------------------------------------
-- AISYSTEM
--
-- Copyright (C) 2012-Present. All rights reserved.
--
-- See LICENSE.TXT document for licensing information.
-------------------------------------------------------------------------------------------------
library ieee;
use ieee.std_logic_1164.all;
use ieee.numeric_std.all;

library ieee_proposed;
use ieee_proposed.fixed_float_types.all;
use ieee_proposed.fixed_pkg.all;
use ieee_proposed.float_pkg.all;

library work;
use work.common.all;

-- gap junction based on conductance
--
-- the gap junction connection model (i.e. electrical synapse) requires two 
-- gap junction objects, one on the source and the other on the target, linked
-- together
--
-- conductance based gap junction models can have conductance G = 0 which is 
-- infinite resistance; in the resistance based models we can not set resistance 
-- R = 0 due to division (1 / R)
--
-- voltage value V is either the target or source object, depending on the object 
-- type that the gap junction is interacting with

entity object8 is  
    port
    (
        reset               : in std_logic;
        clock               : in std_logic;
        -- inputs
        phase           	: in phase_type;
        g_gap               : in float(WE downto -WF);
        v_pre               : in float(WE downto -WF);
        v_post              : in float(WE downto -WF);
        -- outputs
        i_gap               : out float(WE downto -WF)
    );
end entity;

architecture rtl of object8 is
	signal step     : integer range 0 to 31 := 0;
	signal temp1 	: float(WE downto -WF);
begin
	gap_junction : process (reset, clock) is   
    begin
        if reset = '1' then
            i_gap <= (others => '0');
            step <= 0;
        elsif rising_edge(clock) then
            case phase is
                when COMP_PHASE_INITIALIZATION =>
					i_gap <= (others => '0');
                when COMP_PHASE_COMPUTATION =>
                    step <= step + 1;
					case step is
						when 0 =>
							temp1 <= v_pre - v_post;		
						when 1 =>
				            i_gap <= g_gap * temp1;
                        when others => 
                            step <= 0;
					end case;
                when others =>
					null;
			end case;
        end if;
    end process;
end architecture; 

