-------------------------------------------------------------------------------
-- Floating-point sine/cosine (dual operator)
--
-- Generics:
--   - wE [positive] : exponent bit-width
--   - wF [positive] : fraction bit-width
--
-- Ports:
--   - fpX [in]  : operand X in FP format
--   - fpS [out] : sine of X in FP format
--   - fpC [out] : cosine of X in FP format
-------------------------------------------------------------------------------

library ieee;
use ieee.std_logic_1164.all;
use ieee.std_logic_unsigned.all;
library fplib;
use fplib.pkg_common.all;
--use fplib.pkg_fp_common.all;
use fplib.pkg_fp_misc.all;
use fplib.pkg_fp_sin_cos.all;
use fplib.pkg_misc.all;

entity fp_sin_cos_2pi is
  generic ( wE : positive := 6;
            wF : positive := 13 );
  port ( fpX : in  std_logic_vector(wE+wF+2 downto 0);
         fpS : out std_logic_vector(wE+wF+2 downto 0);
         fpC : out std_logic_vector(wE+wF+2 downto 0) );
end entity;

architecture arch of fp_sin_cos_2pi is
  signal eX : std_logic_vector(wE downto 0);
  signal xX : std_logic_vector(1 downto 0);
  signal mX : std_logic_vector(wF downto 0);
  
  signal nY0 : std_logic_vector(wF+2**wE-1 downto 0);
  signal nY1 : std_logic_vector(wF+2 downto 0);
  signal nY  : std_logic_vector(wF+2 downto 0);

  signal far_nZ0 : std_logic_vector(wF-1 downto 0);
  signal far_nZ  : std_logic_vector(wF downto 0);
  signal far_eZ0 : std_logic_vector(log2(wF)-1 downto 0);
  signal far_eZ  : std_logic_vector(wE-1 downto 0);
  signal far_mZ0 : std_logic_vector(wF-1 downto 0);
  signal far_mZ  : std_logic_vector(wF downto 0);

  signal cls_nZ0 : std_logic_vector(wF+2**(wE-1)-1 downto 0);
  signal cls_nZ  : std_logic_vector(wF downto 0);
  signal cls_eZ0 : std_logic_vector(wE-1 downto 0);
  signal cls_eZ  : std_logic_vector(wE-1 downto 0);
  signal cls_mZ  : std_logic_vector(wF downto 0);

  signal nZ : std_logic_vector(wF downto 0);
  signal eZ : std_logic_vector(wE-1 downto 0);
  signal mZ : std_logic_vector(wF downto 0);

  constant nPio4 : std_logic_vector(wF+3 downto 0) := fp_sin_cos_Pio4(wF-1);
  
  signal nS0 : std_logic_vector(wF+1 downto 0);
  signal nS1 : std_logic_vector(wF+3 downto 0);
  signal mS2 : std_logic_vector(2*wF+4 downto 0);
  signal mS3 : std_logic_vector(wF+3 downto 0);
  signal mS4 : std_logic_vector(wF+2 downto 0);
  signal eS4 : std_logic_vector(wE downto 0);
  signal mS5 : std_logic_vector(wF downto 0);
  signal eS5 : std_logic_vector(wE downto 0);
  signal xS  : std_logic_vector(1 downto 0);
  signal eS  : std_logic_vector(wE-1 downto 0);
  signal fS  : std_logic_vector(wF-1 downto 0);

  signal nC0 : std_logic_vector(wF+1 downto 0);
  signal mC1 : std_logic_vector(wF+2 downto 0);
  signal mC2 : std_logic_vector(wF+2 downto 0);
  signal mC3 : std_logic_vector(wF downto 0);
  signal eC  : std_logic_vector(wE-1 downto 0);
  signal fC  : std_logic_vector(wF-1 downto 0);

begin --------------------------------------------------------- Range reduction

  eX <= ("0" & fpX(wE+wF-1 downto wF)) - ("00" & (wE-2 downto 2 => '1') & "00");

  mX <= "1" & fpX(wF-1 downto 0);
  
  shift_ny0 : left_shift
    generic map ( w => wF+1,
                  n => wE )
    port map ( i => mX,
               s => eX(wE-1 downto 0),
               o => nY0 );

  nY1 <= nY0(wF+2 downto 0);

  nY <= nY1  when eX(wE) = '0'   else
        "00" & mX;

  --------------------------------------------------- Range reduction, far path
  
  far_nZ0 <= nY(wF-1 downto 0)                         when nY(wF) = '0' else
             (wF-1 downto 0 => '0') - nY(wF-1 downto 0);
  far_nZ <= far_nZ0 & "0";

  lzc_norm_far_nz : lzc_norm
    generic map ( w => wF,
                  n => log2(wF) )
    port map ( i => far_nZ0,
               z => far_eZ0,
               o => far_mZ0 );

  far_eZ <= ((wE-1 downto log2(wF) => '0') & far_eZ0) +
            ((wE-1 downto 1 => '0') & "1");

  far_mZ <= far_mZ0 & "0";

  ------------------------------------------------- Range reduction, close path

  cls_mZ <= nY(wF downto 0);

  cls_eZ <= (wE-1 downto 0 => '0') - eX(wE-1 downto 0);
  
  shift_cls_nz : right_shift
    generic map ( w => wF+1,
                  n => wE-1 )
    port map ( i => nY(wF downto 0),
               s => cls_eZ(wE-2 downto 0),
               o => cls_nZ0 );

  cls_nZ <= cls_nZ0(wF-1+2**(wE-1)-1 downto -1+2**(wE-1)-1) + ((wF downto 1 => '0') & cls_nZ0(-2+2**(wE-1)-1));

  ----------------------------------------------------- Range reduction, select

  nZ <= far_nZ when eX(wE-1) = '0' else
        cls_nZ;

  eZ <= far_eZ when eX(wE-1) = '0' else
        cls_eZ;

  mZ <= far_mZ when eX(wE-1) = '0' else
        cls_mZ;

  ------------------------------------------------------------------------ Sine

  fsin_nz : fp_sin_cos_fsin
    generic map ( wF => wF-1 )
    port map ( x => nZ,
               r => nS0 );

  nS1 <= nPio4 - ("00" & nS0);

  mS2 <= nS1 * mZ;
  
  mS3(wF+3 downto 1) <= mS2(2*wF+4 downto wF+2);
  mS3(0) <= '0' when mS2(wF+1 downto 0) = (wF+1 downto 0 => '0') else
            '1';

  mS4 <= mS3(wF+3 downto 2) & (mS3(1) or mS3(0)) when mS3(wF+3) = '1' else
         mS3(wF+2 downto 0);

  eS4 <= ("00" & (wE-2 downto 1 => '1') & mS3(wF+3)) - ("0" & eZ);

--   round_sin : fp_round
--     generic map ( wE => wE,
--                   wF => wF )
--     port map ( eX => eS4,
--                mX => mS4,
--                eR => eS5,
--                mR => mS5 );

  round_sin : FP_Round
    generic map ( wE => wE,
                  wF => wF )
    port map ( eA => eS4,
               fA => mS4,
               eR => eS5,
               fR => mS5 );

  xS <= "00" when nZ = (wF downto 0 => '0') and eX(wE) = '0' else
        "01";
  eS <= eS5(wE-1 downto 0);
  fS <= mS5(wF-1 downto 0);

  ---------------------------------------------------------------------- Cosine

  fcos_nz : fp_sin_cos_fcos
    generic map ( wF => wF-1 )
    port map ( x => nZ,
               r => nC0 );

  mC1 <= ("1" & (wF+1 downto 0 => '0')) - ("0" & nC0);

  mC2 <= mC1(wF+2 downto 0) when mC1(wF+2) = '1' else
         mC1(wF+1 downto 0) & "0";

  mC3 <= mC2(wF+2 downto 2) + ((wF downto 1 => '0') & (mC2(1) and (mC2(2) or mC2(0))));

  eC <= "0" & (wE-2 downto 1 => '1') & (mC1(wF+2) or not mC3(wF));
  fC <= mC3(wF-1 downto 0);

  -----------------------------------------------------------------------------

  xX <= fpX(wE+wF+2 downto wE+wF+1);

  fpS(wE+wF+2 downto wE+wF+1) <= "00" when xX = "00" else
                                 "11" when xX(1) = '1' else
                                 xS   when nY(wF) = nY(1+wF) else
                                 "01";

  fpS(wE+wF) <= fpX(wE+wF) when eX(wE) = '1' else
                fpX(wE+wF) xor nY(2+wF);

  fpS(wE+wF-1 downto 0) <= eS & fS when nY(wF) = nY(1+wF) or eX(wE) = '1' else
                           eC & fC;

  fpC(wE+wF+2 downto wE+wF+1) <= "11" when xX(1) = '1' else
                                 "01"  when nY(wF) = nY(1+wF) else
                                 xS;

  fpC(wE+wF) <= fpX(wE+wF) when eX(wE) = '1' else
                fpX(wE+wF) xor nY(2+wF) xor nY(1+wF);

  fpC(wE+wF-1 downto 0) <= eC & fC when nY(wF) = nY(1+wF) or eX(wE) = '1' else
                           eS & fS;

end architecture; -------------------------------------------------------------
