--------------------------------------------------------------------------
-- AISYSTEM
--
-- Copyright (C) 2012-Present. All rights reserved.
--
-- See LICENSE.TXT document for licensing information.
--------------------------------------------------------------------------
library ieee;
use ieee.std_logic_1164.all;
use ieee.numeric_std.all;
use ieee.math_real.all;

library ieee_proposed;
use ieee_proposed.fixed_float_types.all;
use ieee_proposed.fixed_pkg.all;
use ieee_proposed.float_pkg.all;

library work;
use work.common.all;

-- integration test

entity integration_tb is
end;

architecture bench of integration_tb is
	component integration  
		port 
		(
			reset               : in std_logic;
			clock               : in std_logic;
			-- inputs
			req                 : in std_logic;
			dt                  : in unsigned(WEF downto 0);
			y                   : in float(WE downto -WF);
			a                   : in float(WE downto -WF);
			b                   : in float(WE downto -WF);
			-- outputs
			result              : out float(WE downto -WF);
			acknowledge			: out std_logic
		);
	end component;

	constant CLOCK_PERIOD		: time := 10 ns;

	signal reset				: std_logic;
	signal clock				: std_logic;
	signal method				: method_type;
	signal dt					: unsigned(WEF downto 0);
	signal y					: float(WE downto -WF);
	signal a					: float(WE downto -WF);
	signal b					: float(WE downto -WF);
	signal result				: float(WE downto -WF) ;
	signal req					: std_logic := '0';
	signal acknowledge 			: std_logic := '0';
	signal calc_result_f 		: float(WE downto -WF);
	signal is_stopped 			: boolean;
	-- non-synthesizable signals
	signal y_real				: real;
	signal a_real 				: real;
	signal b_real, result_real 	: real;
	signal calc_result 			: real;
	signal k1_r, k2_r 			: real;
	signal k3_r, k4_r 			: real;
begin
    uut : integration
	    port map 
		(
			reset  => reset,
			clock  => clock,
			req    => req,
			dt     => dt,
			y      => y,
			a      => a,
			b      => b,
			result => result,
			acknowledge	=> acknowledge
		);

    y_real <= to_real(y);
    a_real <= to_real(a);
    b_real <= to_real(b);
    result_real <= to_real(result);
    
    stimulus : process
        variable k1, k2, k3, k4, d : real;
        variable dt_r, a_r, b_r, y_r : real;
    begin
        reset <= '1';
        a_r := 0.0;
        b_r := 0.0;
        y_r := 0.0;
        dt_r := 0.0;
        calc_result <= 0.0;
        wait for 5 ns;

        reset <= '0';
        wait for 5 ns;

        -- test bench stimulus code
        wait until rising_edge(clock);
        req <= '1';

        -- select the method to simulate
        method <= INTG_RUNGE_KUTTA4;
        --method <= INTG_FORWARD_EULER;
        --method <= INTG_EXPONENTIAL_EULER;
        --method <= INTG_BACKWARD_EULER;
        --method <= INTG_TRAPEZOID;
        --method <= INTG_CRANK_NICHOLSON;

        for dt_l in 1 to 10 loop
            for a_l in 1 to 5 loop
                for y_l in 1 to 5 loop
                    for b_l in 6 to 10 loop
                        a_r := real(a_l) / 2.0;
                        b_r := real(b_l) / 2.0;
                        y_r := real(y_l) / 2.0;
						
                        dt_r := real(dt_l);
						
                        A <= to_float(a_r, A);
                        B <= to_float(b_r, A);
                        Y <= to_float(y_r, A);
						
                        dt <= to_unsigned(to_float(dt_l, Y), dt);
                        calc_result <= 0.0;
                        wait until rising_edge(clock);

                        case method is
                            when INTG_RUNGE_KUTTA4 =>
                                K1 := (a_r - b_r * y_r) * dt_r;
                                K2 := (a_r - b_r * (y_r + K1 / 2.0)) * dt_r;
                                K3 := (a_r - b_r * (y_r + K2 / 2.0)) * dt_r;
                                K4 := (a_r - b_r * (y_r + K3)) * dt_r;
                                calc_result <= y_r + (K1 + 2.0 * K2 + 2.0 * K3 + K4) / 6.0;

                                K1_r <= K1;
                                K2_r <= K2;
                                K3_r <= K3;
                                K4_r <= K4;
                            when INTG_FORWARD_EULER =>
                                calc_result <= y_r + (a_r - b_r * y_r) * dt_r;
                            when INTG_EXPONENTIAL_EULER =>
                                D := exp(- b_r * dt_r);
                                calc_result <= y_r * D + (a_r / b_r) * (1.0 - D);
                            when INTG_BACKWARD_EULER =>
                                calc_result <= (y_r + a_r * dt_r) / (1.0 + b_r * dt_r);
                            when INTG_TRAPEZOID =>
                                calc_result <= (y_r * (1.0 - dt_r * b_r / 2.0) + dt_r * a_r) / (1.0 + dt_r * b_r / 2.0);    
                            when INTG_CRANK_NICHOLSON =>
                                calc_result <= 2.0 * ((y_r + a_r * dt_r) / (1.0 + b_r * dt_r)) - y_r;
                            when others =>
                                null;
                        end case;
                        wait until acknowledge ='1';

                        assert to_real(result) = calc_result 
							report "ERROR: method " & integer'image(method) & " expected value=" & real'image(calc_result) & " actual value=" & real'image(to_real(result)) & " difference=" & real'image(to_real(result) - calc_result) severity warning;
                        wait until rising_edge(clock);
                    end loop;    
                end loop;    
            end loop;
        end loop;
		
        is_stopped <= true;
        wait;
    end process;

    generation : process
    begin
        while not is_stopped loop
            clock <= '0', '1' after CLOCK_PERIOD / 2;
            wait for CLOCK_PERIOD;
        end loop;
		
        wait;
    end process;
end architecture;


configuration cfg_integration_tb of integration_tb is
	for bench
		for uut : integration
			-- default configuration
		end for;
	end for;
end cfg_integration_tb;

configuration cfg_integration_runge_kutta4 of integration_tb is
	for bench
		for uut : integration
			use entity work.integration(integration_runge_kutta4);
		end for;
	end for;
end cfg_integration_runge_kutta4;

configuration cfg_integration_forward_euler of integration_tb is
	for bench
		for uut : integration
			use entity work.integration(integration_forward_euler);
		end for;
	end for;
end cfg_integration_forward_euler;

configuration cfg_integration_exponential_euler of integration_tb is
	for bench
		for uut : integration
			use entity work.integration(integration_exponential_euler);
		end for;
	end for;
end cfg_integration_exponential_euler;

configuration cfg_integration_backward_euler of integration_tb is
	for bench
		for uut : integration
			use entity work.integration(integration_backward_euler);
		end for;
	end for;
end cfg_integration_backward_euler;

configuration cfg_integration_trapezoid of integration_tb is
	for bench
		for uut : integration
			use entity work.integration(integration_trapezoid);
		end for;
	end for;
end cfg_integration_trapezoid;

configuration cfg_integration_crank_nicholson of integration_tb is
	for bench
		for uut : integration
			use entity work.integration(integration_crank_nicholson);
		end for;
	end for;
end cfg_integration_crank_nicholson;

