--------------------------------------------------------------------------
-- AISYSTEM
--
-- Copyright (C) 2012-Present. All rights reserved.
--
-- See LICENSE.TXT document for licensing information.
--------------------------------------------------------------------------
library ieee;
use ieee.std_logic_1164.all;
use ieee.numeric_std.all;

library ieee_proposed;
use ieee_proposed.fixed_float_types.all;
use ieee_proposed.fixed_pkg.all;
use ieee_proposed.float_pkg.all;

library fplib;
use fplib.pkg_fplib_std.all;

library work;
use work.common.all;
	 
library work;
use work.definition.all;

-- leaky integrate and fire neuron
--
-- c1 and c2 are constants for exponential Euler integration of v_m
-- tau_m is membrane time constant

entity object242 is
	generic
    (
		CLK_FREQUENCY 		: integer := 100 -- [MHz]
    );    
    port 
    (
        reset               : in std_logic;
        clock               : in std_logic;
        -- inputs
        phase               : in phase_type;
		dt					: in float(WEF downto 0); 
        t_step              : in float(BIT48 downto 0);
        absolute_refractory : in float(WE downto -WF);
        r_m                 : in float(WE downto -WF);
        c_m                 : in float(WE downto -WF);
        v_rest              : in float(WE downto -WF);
        i_syn               : in float(WE downto -WF);
        i_gap               : in float(WE downto -WF);
        i_inj               : in float(WE downto -WF);
        v_thr               : in float(WE downto -WF);
        -- outputs
        v_m                 : out float(WE downto -WF)
    );
end entity;


architecture rtl of object242 is
	constant EXP_DELAY      : integer := 0;
			 
	signal r_m_f 			: float(WE downto -WF);
	signal c_m_f 			: float(WE downto -WF);
	signal v_rest_f  		: float(WE downto -WF);
	signal v_m_f  			: float(WE downto -WF);
	signal i_m_f  			: float(WE downto -WF);
	signal tau_m 			: float(WE downto -WF);
	signal time_constant	: float(WE downto -WF);
	signal c1_f				: float(WE downto -WF);
	signal c2_f    			: float(WE downto -WF);
	signal c1v_m         	: float(WE downto -WF);
	signal c2i          	: float(WE downto -WF);
	signal add1				: float(WE downto -WF);
	signal add2    			: float(WE downto -WF);
	signal i_sum_f      	: float(WE downto -WF);   
	signal delay        	: integer range 0 to 15 := 0;
	signal last_fired   	: unsigned(BIT48 downto 0);
	signal clk_div_cnt  	: signed(WEF downto 0);
	signal dt_f         	: unsigned(WEF downto 0);
	signal dt_f_divided 	: float(WE downto -WF);
	signal clk_en       	: std_logic;
	signal step     		: integer range 0 to 31 := 0;
	-- fplib signals
	signal exp_in			: std_logic_vector(WEF2 downto 0);
	signal exp_result		: std_logic_vector(WEF2 downto 0);
begin
	exponent : entity fplib.exp_clk
    	generic map 
		(
			WE    => WE,
            WF    => WF,
            reg   => false,
            szslk => 0
		)
    	port map 
		(
			na  => exp_in,
        	nr  => exp_result,
            clk => clock
		);
    	
    v_m <= v_m_f;
    r_m_f <= to_float(r_m, r_m_f);
    c_m_f <= to_float(c_m, c_m_f);
    v_rest_f <= v_rest;
    dt_f <= dt;
    
	leaky_neuron : process (reset, clock) is variable exp_var : std_logic_vector(WEF downto 0);
	begin
		if reset = '1' then
            step <= 0;
        elsif rising_edge(clock) then
            case phase is
                when COMP_PHASE_INITIALIZATION =>
                    v_m_f <= v_rest;
                    last_fired <= -absolute_refractory;
                    tau_m <= r_m_f * c_m_f;
					
                    if r_m > 0 then
                        i_m_f <= divide(v_rest_f, r_m_f);
                    else
                        i_m_f <= to_float(0, i_m_f);
                    end if;
                when COMP_PHASE_COMPUTATION =>
                    step <= step + 1;
					case step is
						when 0 =>
		                    dt_f_divided <= to_float(dt_f, dt_f_divided) * ((-1) / 100); 
		                when 1 =>
                    		time_constant <= divide(dt_f_divided, tau_m, round_zero, 0, false, false);
                		when 2 =>
		                    exp_var := to_slv(time_constant);
		                    exp_in(WEF2 downto WEF1) <= b"01";
		                    exp_in(WEF downto 0) <= exp_var(WEF downto 0);
		                when 3 =>
		                    if delay = EXP_DELAY then
		                        delay <= 0;
		                    else						
								-- wait
		                        step <= step - 1;		
		                        delay <= delay + 1;
							end if;
                		when 4 =>
		                    if r_m = 0 or c_m = 0 then
		                        c1_f <= to_float(0, c1_f);
		                    else
		                        c1_f <= to_float(exp_result(WEF downto 0), c1_f);
		                    end if;  
                		when 5 => 
		                    c2_f <= (1 - c1_f) * r_m_f;
                		when 6 => 
		                    if clk_en = '1' then  
		                        if (last_fired + absolute_refractory) >= (t_step + dt / 2) then
		                            v_m_f <= v_rest_f;
		                        else
		                            if v_m_f >= v_thr then
		                                last_fired <= t_step;
		                            end if;
		                        end if;
								
		                        add1 <= i_syn + i_gap;
		                        add2 <= i_inj + i_m_f;
								
		                        c1v_m <= c1_f * v_m_f;
		                    end if;    
		                when 7 =>
                        	i_sum_f <= add1 + add2;
		                when 8 => 
                        	c2i <= c2_f * i_sum_f;
                		when 9 => 
                        	v_m_f <= c1v_m + c2i;
                        when others => 
                            step <= 0;
					end case;
                when others =>
                    null;
            end case;
        end if;
   end process;
end architecture;

