------------------------------------------------------------------------------------------------
-- AISYSTEM
--
-- Copyright (C) 2012-Present. All rights reserved.
--
-- See LICENSE.TXT document for licensing information.
-------------------------------------------------------------------------------------------------
library ieee;
use ieee.std_logic_1164.all;
use ieee.numeric_std.all;

library ieee_proposed;
use ieee_proposed.fixed_float_types.all;
use ieee_proposed.fixed_pkg.all;
use ieee_proposed.float_pkg.all;

library work;
use work.common.all;

library work;
use work.definition.all;

-- * izhikevich spiking neuron (20-in-1)
* 
* artificial neuron model from EM Izhikevich "Simple Model of Spiking Neurons",
* IEEE Transactions On Neural Networks, Vol. 14, No. 6, November 2003 pp 1569-1572
*
* parameters A, B, C, and D are used to produce a wide variety of action potential 
* shapes and firing patterns; the "membrane potential" Vm is determined from the 
* equation:
*
*	dVm/dt = Coeff2 * Vm * Vm + Coeff1 * Vm + Coeff0 - u + I
*
* where I is the total incoming 'current' from injection or other sources, with
* du/dt = A*(B*Vm - u), and the spike reset condition:
*
* if (Vm >= VMAX) then Vm = C; u = u + D
* 
* in a normal neural compartment with a finite membrane area and membrane
* capacitance Cm, the 'I' term in the equation for dVm/dt would be Im/Cm;
* in terms of CM, the membrane capacitance per unit membrane area, 
*
* 	Im/Cm = Idens(Amp/area)* area/[CM*area] = Idens/CM
*
* the synaptic channel or other channel object may be coupled to the spiking neuron in the 
* same way as to a compartment, and sending synaptic current to the spiking neuron, and 
* retrieving the Vm to the synpatic channel
*
* integration is done using second order Taylor's series which is more accurate
* than Crank-Nicholson for long runs with large networks:
*		Vm(new) = Vm + dt * Dvt + 0.5 * dt * dt * (d(Dvt) / dt) 
* similarly for u
*
* the following values are based on izh.mod [mV] and [ms]:
*
*        A        B       C      D       I
*================================================================================
*      0.02      0.2     -65     6      14       % tonic spiking
*      0.02      0.25    -65     6       0.5     % phasic spiking
*      0.02      0.2     -50     2      15       % tonic bursting
*      0.02      0.25    -55     0.05    0.6     % phasic bursting
*      0.02      0.2     -55     4      10       % mixed mode
*      0.01      0.2     -65     8      30       % spike frequency adaptation
*      0.02     -0.1     -55     6       0       % Class 1
*      0.2       0.26    -65     0       0       % Class 2
*      0.02      0.2     -65     6       7       % spike latency
*      0.05      0.26    -60     0       0       % subthreshold oscillations
*      0.1       0.26    -60    -1       0       % resonator
*      0.02     -0.1     -55     6       0       % integrator
*      0.03      0.25    -60     4       0       % rebound spike
*      0.03      0.25    -52     0       0       % rebound burst
*      0.03      0.25    -60     4       0       % threshold variability
*      1         1.5     -60     0     -65       % bistability
*      1         0.2     -60   -21       0       % DAP
*      0.02      1       -55     4       0       % accomodation
*     -0.02     -1       -60     8      80       % inhibition-induced spiking
*     -0.026    -1       -45     0      80       % inhibition-induced bursting    

--

PROP
	in:	
		state
		a
		b
		c	
		d
		u
		v_m
		i_tot
		i_inj
		i_syn
		dvt
		dut
	out:
		v_out
CONSTANTS (generic) 
	#define VMAX	30.0
	#define COEFF0	140
	#define COEFF1	5
	#define COEFF2	0.04
INIT
	State = 0;		/* cell is not firing */
			
	U = B * V_m;
	v_out = C;
COMP
			Itot = I + Iinj + Isyn;

			/* Taylor's second order series */
			Dvt = Vm * (COEFF2 * Vm + COEFF1) + COEFF0 - U + Itot;
			Dut = A * (B * Vm - U);
			Vm += Dt * Dvt * (1 + Dt * (COEFF2 * Vm + COEFF1 / 2)) - Dt * Dt * Dut / 2;
			U += Dt * Dut + Dt * Dt * A * (B * Dvt - Dut);

			if (Vm >= VMAX)
			{
				/* reset the action potential to its falling side, when it passes maximum voltage */
				State = 1;

				Vm = C;
				U += D;
			}
	
			v_out = v_m


entity objectX is  
    port
    (
        reset               : in std_logic;
        clock               : in std_logic;
        -- inputs
        phase           	: in phase_type;
        prop1	            : in float(WE downto -WF);
        prop2	            : in float(WE downto -WF);
        prop3	            : in float(WE downto -WF);
        -- outputs
        prop4	            : out float(WE downto -WF);
    );
end entity;

architecture rtl of objectX is
	signal step     : integer range 0 to 31 := 0;
	signal temp1 	: float(WE downto -WF);
begin
	object_name : process (reset, clock) is   
    begin
        if reset = '1' then

			step <= 0;
        elsif rising_edge(clock) then
            case phase is
                when COMP_PHASE_INITIALIZATION =>
					
                when COMP_PHASE_COMPUTATION =>
                    step <= step + 1;
					case step is
						when 0 =>
							
						when 1 =>
				            
                        when others => 
                            step <= 0;
					end case;
                when others =>
					null;
			end case;
        end if;
    end process;
end architecture; 

